qss = f'''
#Popup {{
    background-color: rgba(255, 255, 255, 0.1);
}}

#PopupFrame {{
    background-color: #17212B;
    width: 400px;
    height: 200px;
    border-radius: 20px;
}}

#PopupMessageLbl {{
    color: white;
    font-size: 24px;
    background-color: transparent;
}}

#PopupOkBtn,
#PopupCancelBtn,
#PopupYesBtn,
#PopupNoBtn {{
    color: white;
    border: none;
    font-size: 18px; 
    border-radius: 5px;
    min-height: 40px;
}}

#PopupOkBtn,
#PopupYesBtn {{
    background-color: darkgreen;
}}

#PopupOkBtn:hover,
#PopupYesBtn:hover {{
    background-color: rgba(0, 255, 0, 0.5);
}}

#PopupCancelBtn,
#PopupNoBtn {{
    background-color: rgba(182, 0, 40, 1);
}}

#PopupCancelBtn:hover,
#PopupNoBtn:hover {{
    background-color: rgba(182, 0, 40, 0.5);
}}
'''



from PySide6.QtWidgets import QWidget, QCompleter, QStyledItemDelegate
from PySide6.QtCore import Qt
from typing import Sequence

from ..input import LineInput


class SearchBar(LineInput):
    class Completer(QCompleter):
        def __init__(self, parent: QWidget, *, items: Sequence[str], qss: str = ''):
            super().__init__(items, parent)
            self.setCaseSensitivity(Qt.CaseInsensitive)
            self.popup().setObjectName(f'{parent.objectName()}Popup')
            self.popup().setItemDelegate(QStyledItemDelegate(self))
            self.popup().setStyleSheet(qss)

    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        LineInput.__init__(self, parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            completer: Completer,
            **kwargs
    ) -> 'LineInput':
        await LineInput.init(self, **kwargs)
        self.setCompleter(completer)
        return self



from PySide6.QtWidgets import QPushButton

from ..types import Parent, QSS
from ..context import ContextObj


class Button(ContextObj, QPushButton):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QPushButton.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self,
            **kwargs
    ) -> 'Button':
        return await self._render(**kwargs)



from .state_button import StateButton
from .image_button import ImageButton
from .total_button import TotalButton
from .searchbar import SearchBar
from .duration_label import DurationLabel
from .popup import Popup



from PySide6.QtWidgets import QFrame, QLayout

from ..types import QSS, Parent
from ..context import ContextObj


class Frame(ContextObj, QFrame):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QFrame.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            style: ... = None, layout: QLayout = None,
            **kwargs,
    ) -> 'Frame':
        if style:
            self.setFrameStyle(style)
        if layout:
            self.setLayout(layout)
        return await self._render(**kwargs)



from PySide6.QtGui import QResizeEvent
from qasync import asyncSlot
from typing import Iterable

from ..frame import Frame
from ..layout import Layout
from ..button import Button
from ..label import Label
from ...types import Parent, Event, DefaultEvent, QSS, LogicButton


class Popup(LogicButton, Frame):
    def __init__(
            self, parent: Parent, name: str = None, qss: QSS = None,
            *,
            message: str = '',
            buttons: Iterable[LogicButton] = (LogicButton.YES, LogicButton.NO),
            on_success: Event = DefaultEvent,
            on_failure: Event = DefaultEvent
    ):
        super().__init__(parent, name if name else self.__class__.__name__, qss)
        self.message: str = message
        self.buttons: Iterable[LogicButton] = buttons
        self.on_success: Event = on_success
        self.on_failure: Event = on_failure
        self.hide()

    @asyncSlot()
    async def display(self) -> None:
        self.setLayout(await Layout.vertical().init(
            spacing=10, alignment=Layout.Center, margins=(20, 20, 20, 20),
            items=[
                await Frame(self, f'{self.objectName()}Frame').init(
                    layout=await Layout.vertical().init(
                        spacing=20, margins=(20, 20, 20, 20),
                        items=[
                            await Label(self, f'{self.objectName()}MessageLbl').init(
                                text=self.message, wrap=True, alignment=Layout.Center
                            ),
                            await Layout.horizontal().init(
                                spacing=20,
                                items=[
                                    await Button(self, f'{self.objectName()}YesBtn', Popup.YES in self.buttons).init(
                                        text='Yes', on_click=lambda: self.__event(self.on_success)
                                    ),
                                    await Button(self, f'{self.objectName()}NoBtn', Popup.NO in self.buttons).init(
                                        text='No', on_click=lambda: self.__event(self.on_failure)
                                    ),
                                    await Button(self, f'{self.objectName()}OkBtn', Popup.OK in self.buttons).init(
                                        text='Ok', on_click=lambda: self.__event(self.on_success)
                                    ),
                                    await Button(self, f'{self.objectName()}CancelBtn', Popup.CANCEL in self.buttons).init(
                                        text='Cancel', on_click=lambda: self.__event(self.on_failure)
                                    )
                                ]
                            )
                        ]
                    )
                )
            ]
        ))
        self.show()

    @asyncSlot()
    async def __event(self, event: Event):
        self._emit(event)
        self.deleteLater()

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.resize(self.parent().size())



from PySide6.QtWidgets import QSplitter
from typing import Iterable

from .frame import Frame
from .extensions import SplitterWidgetExt
from ..context import ContextObj
from ..types import QSS, Parent, Orientation


class Splitter(ContextObj, Orientation, QSplitter):
    class Handle(Frame):
        def __init__(self, parent: Parent, name: str):
            super().__init__(parent, name, True)

    class Widget(SplitterWidgetExt, Frame):
        def __init__(self, parent: Parent, name: str, *, collapsible: bool = True,
                     expand_to: int, expand_min: int = None, expand_max: int = None):
            Frame.__init__(self, parent, name, True)
            SplitterWidgetExt.__init__(self, expand_to, expand_min, expand_max, collapsible)

    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None,
                 orientation: Orientation.Orientation = Orientation.Horizontal):
        QSplitter.__init__(self, orientation, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            items: Iterable['Splitter.Widget'] = (),
            **kwargs
    ) -> 'Splitter':
        for item in items:
            self.addWidget(item)
        return await self._render(**kwargs)

    def addWidget(self, widget: 'Splitter.Widget') -> None:
        super().addWidget(widget)
        widget.splitter = self
        widget.orientation = self.orientation()
        self.setCollapsible(self.count() - 1, widget.collapsible)



# to be used together with other imports
from ..types import Parent
# exist normally, but added to keep context completed
from .window import Window
from .statusbar import StatusBar
from .panel import Panel
# standard frame-like
from .widget import Widget
from .frame import Frame
# specific frame-like
from .stacked_widget import StackedWidget
from .splitter import Splitter
from .scroll_area import ScrollArea
# modular
from .button import Button
from .label import Label
from .input import Input
from .selector import Selector
from .slider import Slider
from .date_time import DateTime
# layout
from .layout import Layout
from .spacer import Spacer



from PySide6.QtGui import QFontMetrics, QPaintEvent
from PySide6.QtWidgets import QLabel

from ..types import Parent, QSS, Alignment, ElideMode, SizePolicy
from ..context import ContextObj


class Label(ContextObj, Alignment, SizePolicy, ElideMode, QLabel):
    __elide_mode: ElideMode.ElideMode
    __pre_elilded_text: str = None

    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QLabel.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            wrap: bool = False, elide: ElideMode.ElideMode = None,
            **kwargs
    ) -> 'Label':
        self.__elide_mode = elide
        self.setWordWrap(wrap)
        return await self._render(**kwargs)

    def paintEvent(self, event: QPaintEvent) -> None:
        if self.is_elided:
            text = self.__pre_elilded_text
            self.setText(QFontMetrics(self.font()).elidedText(self.__pre_elilded_text, self.__elide_mode, self.width() - 10))
            self.__pre_elilded_text = text
        QLabel.paintEvent(self, event)

    @property
    def is_elided(self) -> bool:
        return self.__elide_mode is not None

    def setText(self, text: str) -> None:
        self.__pre_elilded_text = text
        QLabel.setText(self, text)



from PySide6.QtWidgets import QWidget, QLayout

from ..context import ContextObj
from ..types import QSS, Parent


class Widget(ContextObj, QWidget):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QWidget.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            layout: QLayout = None,
            **kwargs
    ) -> 'Widget':
        if layout:
            self.setLayout(layout)
        return await self._render(**kwargs)



from PySide6.QtWidgets import QMessageBox


class LogicButton:
    LogicButton = QMessageBox.StandardButton
    YES = QMessageBox.StandardButton.Yes
    NO = QMessageBox.StandardButton.No
    OK = QMessageBox.StandardButton.Ok
    CANCEL = QMessageBox.StandardButton.Cancel



from PySide6.QtWidgets import QStatusBar

from ..context import ContextObj
from ..types import Parent


class StatusBar(ContextObj, QStatusBar):
    def __init__(self, parent: Parent, name: str, visible: bool = True):
        QStatusBar.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)



from PySide6.QtWidgets import QWidget, QLayoutItem, QSpacerItem
from PySide6.QtCore import Qt, QObject
from typing import Sequence, Iterable

from ...types import Alignment


class LayoutExt(Alignment):
    async def init(
            self, *,
            items: Sequence[QObject] = (),
            **kwargs
    ) -> 'LayoutExt':
        # add items according to the rule item=`widget` or items=`widget, alignment` (not a tuple, but sequentially)
        i = 0
        while i < len(items):
            if i + 1 < len(items) and isinstance(items[i + 1], (Qt.AlignmentFlag, Qt.Alignment)):
                self.add(items[i], items[i + 1])
                i += 1
            else:
                self.add(items[i])
            i += 1
        # set defaults through kwargs if values are not specified
        if not kwargs.get('margins'):
            kwargs['margins'] = (0, 0, 0, 0)
        if not kwargs.get('spacing'):
            kwargs['spacing'] = 0
        return await self._render(**kwargs)  # from ContextObj

    def add(self, obj: QObject, alignment: Alignment.Alignment = None) -> QObject:
        if isinstance(obj, QWidget):
            if alignment:
                self.addWidget(obj, alignment=alignment)
            else:
                self.addWidget(obj)
        elif isinstance(obj, QSpacerItem):
            self.addSpacerItem(obj)
        elif isinstance(obj, QLayoutItem):
            self.addLayout(obj)
        else:
            raise TypeError(f'{obj} has unsupported type {type(obj)} to `add` to {self}')
        return obj

    def reorder(self, at: int, to: int):
        self.removeItem(item := self.itemAt(at))
        self.insertItem(to, item)

    def clear(self, exceptions: Iterable[QObject] = ()):
        for i in reversed(range(self.count())):
            item = self.itemAt(i).widget()
            if not item:
                item = self.itemAt(i).layout()
            if not item:
                item = self.itemAt(i).spacerItem()
            if item not in exceptions:
                item.setParent(None)



from . import colors

qss = f'''
#SearchBar {{
    background-color: {colors.LAYER_3};
    font-size: 18px;
    padding: 5px;
    color: {colors.TEXT_PRIMARY};
    border: none;
}}

#SearchBarPopup {{
    background-color: {colors.LAYER_1};
    border: none;
    font-size: 20px;
    font-weight: bold;
    color: {colors.TEXT_PRIMARY};
}}

QListView::item:hover {{
    background-color: {colors.LAYER_5};
}}
'''



from PySide6.QtWidgets import QSpacerItem

from ..types import SizePolicy


class Spacer(SizePolicy, QSpacerItem):
    def __init__(
            self,
            hpolicy: SizePolicy.SizePolicy = SizePolicy.Minimum,
            vpolicy: SizePolicy.SizePolicy = SizePolicy.Minimum
    ):
        QSpacerItem.__init__(self, 0, 0, hpolicy, vpolicy)



from PySide6.QtWidgets import QStackedWidget, QFrame
from PySide6.QtCore import QObject
from typing import Iterable

from ..context import ContextObj
from ..types import QSS, Parent, Alignment, Orientation


class StackedWidget(ContextObj, Orientation, QStackedWidget):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QStackedWidget.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

        # workaround which includes override of `setCurrentIndex`, otherwise problems with `parent()`
        self.addWidget(QFrame(self))  # QFrame is likely the most lightweight QWidget in Qt6

    async def init(
            self, *,
            alignment: Alignment.Alignment = Alignment.HCenter, items: Iterable[QObject] = (),
            **kwargs
    ) -> 'StackedWidget':
        self.layout().setAlignment(alignment)
        for item in items:
            self.addWidget(item)
        return await self._render(**kwargs)

    def setCurrentIndex(self, index: int) -> None:
        super().setCurrentIndex(index + 1)



from PySide6.QtWidgets import QLineEdit, QTextEdit, QPlainTextEdit
from abc import ABC

from ..types import QSS, Parent, EchoMode
from .extensions import InputExt
from ..context import ContextObj


class LineInput(ContextObj, InputExt, QLineEdit):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QLineEdit.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self, *,
            hidden: bool = False,
            **kwargs
    ) -> 'LineInput':
        if hidden:
            self.hide_echo()
        return await self._render(**kwargs)

    def hide_echo(self):
        self.setEchoMode(EchoMode.Hidden)

    def show_echo(self):
        self.setEchoMode(EchoMode.Default)

    def toggle_echo(self):
        self.hide_echo() if self.echoMode() == EchoMode.Default else self.show_echo()


class ReachInput(ContextObj, InputExt, QTextEdit):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QTextEdit.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self,
            **kwargs
    ) -> 'ReachInput':
        return await self._render(**kwargs)


class PlainInput(ContextObj, InputExt, QPlainTextEdit):
    def __init__(self, parent: Parent, name: str, visible: bool = True, qss: QSS = None):
        QPlainTextEdit.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.qss = qss

    async def init(
            self,
            **kwargs
    ) -> 'PlainInput':
        return await self._render(**kwargs)


class Input(ABC, EchoMode):
    @staticmethod
    def line(parent: Parent, name: str, visible: bool = True, qss: QSS = None) -> LineInput:
        return LineInput(parent, name, visible, qss)

    @staticmethod
    def reach(parent: Parent, name: str, visible: bool = True, qss: QSS = None) -> ReachInput:
        return ReachInput(parent, name, visible, qss)

    @staticmethod
    def plain(parent: Parent, name: str, visible: bool = True, qss: QSS = None) -> PlainInput:
        return PlainInput(parent, name, visible, qss)



from . import colors

qss: str = f'''
#DateTime {{
    color: white;
    background-color: black;
    padding: 0px;
    border: none;
    font-size: 18px;
}}

QCalendarWidget QToolButton {{
    height: 30px;
    color: white;
    min-width: 75px;
    icon-size: 20px 20px;
    background-color: {colors.TRANSPARENT};
    border: none;
}}

QCalendarWidget QMenu {{ /* drop-down month selection */
    color: white;
    font-size: 18px;
    background-color: gray;
}}

QCalendarWidget QSpinBox {{ /* up and down buttons frame of changing year */
    width: 50px; 
    font-size:24px; 
    color: white; 
    background-color: {colors.TRANSPARENT};
}}

QCalendarWidget QSpinBox::up-button,
QCalendarWidget QSpinBox::down-button {{
    width: 25px;
    height: 18px;
}}

QCalendarWidget QWidget {{  /* row with day names */
    alternate-background-color: gray;
    font-size: 16px;
}}

QCalendarWidget QWidget#qt_calendar_navigationbar {{ /* most top frame */
    background-color: black;
}}

QCalendarWidget QAbstractItemView:enabled,
QCalendarWidget QAbstractItemView:disabled {{
    font-size: 16px;
}}

QCalendarWidget QAbstractItemView:enabled {{ /* days in this months */
    background-color: {colors.TRANSPARENT};
    selection-background-color: black;
    selection-color: rgb(0, 255, 0);
    color: white; 
}}

QCalendarWidget QAbstractItemView:disabled {{ /* days in other months */
    color: rgb(64, 64, 64); 
}}
'''
